from tuned_lens.residual_stream import record_residual_stream
from transformers import GPT2LMHeadModel, GPT2Config
import torch as th


def test_record_residual_stream():
    """
    Test that every other residual stream state is identical to the corresponding
    hidden state reported by HuggingFace.
    """
    model = GPT2LMHeadModel(GPT2Config())

    with record_residual_stream(model, sublayers=True) as stream:
        outputs = model(th.ones(1, 1, dtype=th.long), output_hidden_states=True)

    # Exclude the final hidden state because HF applies ln_f while we don't
    their_hiddens = outputs.hidden_states[1:-1]
    for their_hidden, our_hidden in zip(their_hiddens, stream.layers):
        th.testing.assert_close(their_hidden, our_hidden)



from .causal import ablate_layer, layer_intervention
from .nn import TunedLens
from .plotting import plot_lens
from .stats import LogitStats, ResidualStats
from .residual_stream import record_residual_stream, ResidualStream



from tuned_lens.nn.lenses import TunedLens, LogitLens
import transformers as trf

import tempfile
import torch as th
import pytest
import mock


@pytest.fixture
def logit_lens():
    model = mock.MagicMock(trf.PreTrainedModel)
    model.config = mock.MagicMock(trf.PretrainedConfig)
    model.config.hidden_size = 128
    model.config.num_layers = 3
    model.config.vocab_size = 100
    model.get_output_embeddings = mock.MagicMock(return_value=th.nn.Linear(128, 100))

    with mock.patch("tuned_lens.model_surgery.get_final_layer_norm") as get_final_ln:
        get_final_ln.return_value = th.nn.LayerNorm(128)

    return LogitLens(model)


@pytest.fixture
def tuned_lens():
    return TunedLens(d_model=128, num_layers=3, vocab_size=100)


def test_logit_lens_smoke(logit_lens):
    randn = th.randn(1, 10, 128)
    logit_lens(randn, 0)


def test_tuned_lens_smoke(tuned_lens: TunedLens):
    randn = th.randn(1, 10, 128)
    logits_forward = tuned_lens(randn, 0)
    logits = tuned_lens.unembedding(
        tuned_lens.layer_norm(randn + tuned_lens[0](tuned_lens.layer_norm(randn)))
    )
    assert th.allclose(logits_forward, logits)


def test_tuned_lens_save_and_load(tuned_lens):
    randn = th.randn(1, 10, 128)

    logits_before = tuned_lens(randn, 1)
    with tempfile.TemporaryDirectory() as tmpdir:
        tuned_lens.save(tmpdir)
        tuned_lens = TunedLens.load(tmpdir)
        logits_after = tuned_lens(randn, 1)
        assert th.allclose(logits_before, logits_after)



from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import starmap, zip_longest

from .model_surgery import get_transformer_layers
from typing import Callable, Generator, Iterable, Optional, overload, Type, Union
import torch as th
import torch.distributed as dist


@dataclass
class ResidualStream:
    """Collection of transformer hidden states, or derived quantities."""

    embeddings: Optional[th.Tensor] = None
    attentions: list[th.Tensor] = field(default_factory=list)
    layers: list[th.Tensor] = field(default_factory=list)

    @classmethod
    def stack(cls, streams: list["ResidualStream"]) -> "ResidualStream":
        """Stack a list of residual streams into a single stream."""
        if len(streams) < 2:
            raise ValueError("Expected at least two streams to stack")

        first, *rest = streams
        return first.zip_map(lambda *tensors: th.stack(tensors), *rest)

    def all_reduce_(self):
        """All-reduce all states across workers if needed."""
        if not dist.is_initialized():
            return

        for state in self:
            dist.all_reduce(state)
            state /= dist.get_world_size()

    def clear(self) -> None:
        """Clear all residual states."""
        self.embeddings = None
        self.attentions.clear()
        self.layers.clear()

    @property
    def has_sublayers(self) -> bool:
        """Return whether the stream contains both attention and layer outputs."""
        return bool(self.attentions) and bool(self.layers)

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the first state."""
        return next(iter(self)).shape

    def items(
        self, reverse: bool = False
    ) -> Generator[tuple[str, th.Tensor], None, None]:
        """Iterate over residual states in topological order, yielding labels."""
        # Forward order
        if not reverse:
            if self.embeddings is not None:
                yield "input", self.embeddings

            # Yield attention and layer outputs in alternating order
            for i, (attn, layer) in enumerate(
                zip_longest(self.attentions, self.layers)
            ):
                if attn is not None:
                    yield f"{i}.attn", attn
                if layer is not None:
                    yield f"{i}.ffn", layer

        # Reverse order
        else:
            # Yield attention and layer outputs in alternating order
            attns, layers = reversed(self.attentions), reversed(self.layers)
            for i, (attn, layer) in enumerate(zip_longest(attns, layers)):
                if layer is not None:
                    yield f"{i}.ffn", layer
                if attn is not None:
                    yield f"{i}.attn", attn

            if self.embeddings is not None:
                yield "input", self.embeddings

    def labels(self) -> list[str]:
        """Return labels for the residual states suitable for e.g. plotting."""
        return [label for label, _ in self.items()]

    def map(self, fn: Callable, reverse: bool = False) -> "ResidualStream":
        """Map a function over all states, returning a new `ResidualStream`."""
        it = reversed(self) if reverse else iter(self)
        return self.new_from_list(list(map(fn, it)))

    def pairwise_map(
        self, fn: Callable[[th.Tensor, th.Tensor], th.Tensor]
    ) -> "ResidualStream":
        """Map over adjacent pairs of states, returning a new `ResidualStream`."""
        if self.embeddings is None:
            raise ValueError("Can't map pairwise without input embeddings")

        states = list(self)
        return self.new_from_list(list(starmap(fn, zip(states[:-1], states[1:]))))

    def zip_map(self, fn: Callable, *others: "Iterable") -> "ResidualStream":
        """Map over corresponding states, returning a new `ResidualStream`."""
        return self.new_from_list(list(starmap(fn, zip(self, *others))))

    def new_from_list(self, states: list[th.Tensor]) -> "ResidualStream":
        """Create a new `ResidualStream` with the given states."""
        embeddings = states.pop(0) if self.embeddings is not None else None

        if self.has_sublayers:
            return ResidualStream(
                embeddings=embeddings, attentions=states[::2], layers=states[1::2]
            )
        else:
            return ResidualStream(embeddings=embeddings, layers=states)

    def plot(self, tick_spacing: int = 2, **kwargs):
        """Plot the residual states."""
        import matplotlib.pyplot as plt

        plt.plot(self, **kwargs)
        if not self.has_sublayers:
            plt.xlabel("Layer")
        else:
            plt.xlabel("Sublayer")
            plt.xticks(
                labels=[
                    label
                    for i, label in enumerate(self.labels())
                    if i % tick_spacing == 0
                ],
                ticks=range(0, len(self), tick_spacing),
                rotation=60,
            )

    def mean_update(self, other: "ResidualStream", i: int) -> "ResidualStream":
        """Interpret `self` as a running mean and update it with `other`."""
        return self.zip_map(lambda mu, x: i / (i + 1) * mu + 1 / (i + 1) * x, other)

    def residuals(self) -> "ResidualStream":
        """Compute residual (hidden state diff) for each block."""
        return self.pairwise_map(lambda s1, s2: s2 - s1)

    def __contains__(self, item: th.Tensor) -> bool:
        """Check if the stream contains a reference to this exact tensor."""
        return any(item.is_set_to(state) for state in self)

    @overload
    def __getitem__(self, item: slice) -> list[th.Tensor]:
        ...

    @overload
    def __getitem__(self, item: int) -> th.Tensor:
        ...

    def __getitem__(self, item: Union[int, slice]):
        """Return the state at the given index, or with the given name."""
        return list(self)[item]

    def __iter__(self) -> Generator[th.Tensor, None, None]:
        """Iterate over residual states in topological order."""
        for _, state in self.items():
            yield state

    def __len__(self) -> int:
        """Return the number of residual states."""
        num_states = len(self.attentions) + len(self.layers)
        if self.embeddings is not None:
            num_states += 1

        return num_states

    def __reversed__(self) -> Generator[th.Tensor, None, None]:
        """Iterate over residual states in reverse topological order."""
        for _, state in self.items(reverse=True):
            yield state


@contextmanager
def record_residual_stream(
    model: th.nn.Module,
    *,
    include_input: bool = True,
    norm_class: Type[th.nn.Module] = th.nn.LayerNorm,
    post_norm: bool = False,
    retain_grads: bool = False,
    sublayers: bool = False,
) -> Generator[ResidualStream, None, None]:
    """Record every state of the residual stream in a transformer forward pass.

    This is a context manager that adds forward hooks to each `nn.LayerNorm` module
    in the transformer, storing the input in a dictionary keyed by the layer norm's
    name. This dictionary is yielded by the context manager for later analysis.

    If you want to record multiple forward passes, you should either call `clear()`
    on the stream object or create a new `record_residual_stream` context each time.

    Args:
        model: The transformer model to record.
        include_input: Whether to include the input embeddings in the stream.
        norm_class: The class of normalization layer to record.
        post_norm: Whether the transformer uses LayerNorm after residual connections.
        retain_grads: Whether to retain gradients for the recorded states.
        sublayers: Whether to record attention and layer outputs separately.
    """
    hooks = []
    residual_stream = ResidualStream()

    def process(state: th.Tensor) -> th.Tensor:
        if retain_grads:
            state.requires_grad_(True)
            state.retain_grad()
        else:
            state = state.detach()

        return state

    def store_embeddings(module: th.nn.Module, inputs) -> None:
        residual_stream.embeddings = process(inputs[0])

    def store_attn(module: th.nn.Module, inputs) -> None:
        residual_stream.attentions.append(process(inputs[0]))

    def store_layer(module: th.nn.Module, inputs, output: Union[th.Tensor, tuple]):
        # HuggingFace layers usually return tuples
        if isinstance(output, tuple):
            output = output[0]
            if not isinstance(output, th.Tensor):
                idx = len(residual_stream.layers)
                raise RuntimeError(
                    f"Expected first element of layer {idx} output to be a Tensor"
                )

        residual_stream.layers.append(output)

    _, layers = get_transformer_layers(model)
    if include_input:
        hooks.append(layers[0].register_forward_pre_hook(store_embeddings))

    for i, layer in enumerate(layers):
        hooks.append(layer.register_forward_hook(store_layer))

        # For sublayers=True, we need to hook into one of the layer norms in this layer
        if sublayers:
            layer_norms = [m for m in layer.modules() if isinstance(m, norm_class)]
            if not layer_norms:
                if sublayers:
                    raise ValueError(
                        f"No LNs found in layer {i}; try specifying `norm_class`"
                    )
                else:
                    continue
            elif len(layer_norms) != 2:
                if sublayers:
                    raise ValueError(
                        f"Expected 2 layer norms per layer when sublayers=True, "
                        f"but layer {i} has {len(layer_norms)}"
                    )
                else:
                    continue

            post_attn_ln = layer_norms[0 if post_norm else 1]
            hooks.append(post_attn_ln.register_forward_pre_hook(store_attn))

    try:
        yield residual_stream

    # Make sure we remove the hooks even if an exception is raised.
    finally:
        for hook in hooks:
            hook.remove()



from tuned_lens.causal import remove_subspace
import pytest
import torch as th


@pytest.mark.parametrize("d", list(range(1, 1000, 100)))
def test_remove_subspace(d: int):
    a = th.randn(10, d, dtype=th.float64)

    for k in range(1, d, 10):
        b = th.randn(d, k, dtype=th.float64)
        inner = a @ b

        a_ = remove_subspace(a, b, mode="zero")
        inner_ = a_ @ b
        th.testing.assert_close(inner_, th.zeros_like(inner_))

        a_ = remove_subspace(a, b, mode="mean")
        inner_ = a_ @ b
        th.testing.assert_close(inner_, inner.mean(0, keepdim=True).expand_as(inner_))



from torch.distributions import Dirichlet, kl_divergence
from tuned_lens import ResidualStats, ResidualStream
from tuned_lens.stats import LogitStats
import pytest
import random
import torch as th


def test_logit_stats_correctness():
    """Test that `LogitStats` recovers the true Dirichlet within a small error."""
    th.manual_seed(42)

    x = Dirichlet(th.tensor([1.0, 1.0, 1.0]))
    logits1 = x.sample(th.Size([10000])).log() + random.uniform(-0.1, 0.1)
    logits2 = x.sample(th.Size([10000])).log() + random.uniform(-0.1, 0.1)

    stats = LogitStats()
    stats.update(logits1)
    stats.update(logits2)
    x2 = stats.mle()

    assert kl_divergence(x, x2) < 1e-3


CONFIGS: list[tuple[str, th.dtype]] = [("cpu", th.float32)]
if th.cuda.is_available():
    # Don't require CUDA for tests
    CONFIGS.append(("cuda", th.float16))
    CONFIGS.append(("cuda", th.float32))


@pytest.mark.parametrize("config", CONFIGS)
def test_residual_stats_correctness(config):
    """Test that the `ResidualStats` mean and covariance are correct."""
    device, dtype = config

    th.manual_seed(42)
    N = 100  # Number of batches
    B = 2  # Batch size
    D = 256  # Hidden dimension
    L = 3  # Number of layers
    S = 128  # Sequence length

    # Generate correlated random data from a multivariate Gaussian
    A = th.randn(D, D, device=device) / D**0.5
    master_cov = A @ A.T  # Ensure positive definite
    dist = th.distributions.MultivariateNormal(
        loc=th.randn(D, device=device), covariance_matrix=master_cov
    )
    batch_dims = th.Size([B, S])

    stats = ResidualStats()
    streams = []
    for _ in range(N):
        stream = ResidualStream(
            embeddings=dist.sample(batch_dims).to(dtype),
            attentions=[dist.sample(batch_dims).to(dtype)] * L,
            layers=[dist.sample(batch_dims).to(dtype)] * L,
        )
        streams.append(stream)
        stats.update(stream)

    # Stack all the streams together so we can compute the non-streaming stats
    master = ResidualStream.stack(streams).map(lambda h: h.view(N * B * S, D))

    for h, mu, cov, norm in zip(
        master, stats.mean(), stats.covariance(), stats.mean_norm()
    ):
        th.testing.assert_close(mu.to(h.dtype), h.mean(0))
        th.testing.assert_close(cov.to(h.dtype), h.T.cov(), atol=0.005, rtol=1e-5)
        th.testing.assert_close(norm.to(h.dtype), h.norm(dim=-1).mean())

        # We should at least get close-ish to the true covariance
        assert th.dist(cov, master_cov) / th.norm(master_cov) < 0.1



from .scripts.lens import main as lens_main
from argparse import ArgumentParser
from contextlib import nullcontext, redirect_stdout
from pathlib import Path
import os
import torch.distributed as dist


def run():
    parser = ArgumentParser(
        description="Train or evaluate a set of tuned lenses for a language model.",
    )
    # Arguments shared by train and eval; see https://stackoverflow.com/a/56595689.
    parent_parser = ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "model_name", type=str, help="Name of model to use in the Huggingface Hub."
    )
    parent_parser.add_argument(
        "dataset",
        type=str,
        default=("the_pile", "all"),
        nargs="*",
        help="Name of dataset to use. Can either be a local .jsonl file or a name "
        "suitable to be passed to the HuggingFace load_dataset function.",
    )
    parent_parser.add_argument(
        "--fsdp",
        action="store_true",
        help="Run the model with Fully Sharded Data Parallelism.",
    )
    parent_parser.add_argument(
        "--loss",
        type=str,
        default="kl",
        choices=("ce", "kl"),
        help="Loss function to use.",
    )
    parent_parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Don't permanently cache the model on disk.",
    )
    parent_parser.add_argument(
        "--per-gpu-batch-size",
        type=int,
        default=1,
        help="Number of samples to try to fit on a GPU at once.",
    )
    parent_parser.add_argument(
        "--random-model",
        action="store_true",
        help="Use a randomly initialized model instead of pretrained weights.",
    )
    parent_parser.add_argument(
        "--residual-stats",
        action="store_true",
        help="Save means and covariance matrices for states in the residual stream.",
    )
    parent_parser.add_argument(
        "--revision",
        type=str,
        default="main",
        help="Git revision to use for pretrained models.",
    )
    parent_parser.add_argument(
        "--seed", type=int, default=42, help="Random seed for data shuffling."
    )
    parent_parser.add_argument(
        "--slow-tokenizer", action="store_true", help="Use a slow tokenizer."
    )
    parent_parser.add_argument(
        "--split", type=str, default="validation", help="Split of the dataset to use."
    )
    parent_parser.add_argument(
        "--sweep", type=str, help="Range of checkpoints to sweep over"
    )
    parent_parser.add_argument(
        "--task", type=str, nargs="+", help="lm-eval task to run the model on."
    )
    parent_parser.add_argument(
        "--text-column",
        type=str,
        default="text",
        help="Column of the dataset containing text to run the model on.",
    )
    parent_parser.add_argument(
        "--tokenizer",
        type=str,
        help="Name of pretrained tokenizer to use from the Huggingface Hub. If None, "
        'will use AutoTokenizer.from_pretrained("<model name>").',
    )
    parent_parser.add_argument(
        "--token-shift",
        type=int,
        default=None,
        help="How to shift the labels wrt the input tokens (1 = next token, "
        "0 = current token, -1 = previous token, etc.)",
    )

    subparsers = parser.add_subparsers(dest="command")
    train_parser = subparsers.add_parser("train", parents=[parent_parser])
    downstream_parser = subparsers.add_parser("downstream", parents=[parent_parser])
    eval_parser = subparsers.add_parser("eval", parents=[parent_parser])

    # Training-only arguments
    train_parser.add_argument(
        "--constant", action="store_true", help="Train only the bias term."
    )
    train_parser.add_argument(
        "--extra-layers", type=int, default=0, help="Number of extra decoder layers."
    )
    train_parser.add_argument(
        "--lasso", type=float, default=0.0, help="LASSO (L1) regularization strength."
    )
    train_parser.add_argument(
        "--lens", type=Path, help="Directory containing a lens to warm-start training."
    )
    train_parser.add_argument(
        "--lr-scale",
        type=float,
        default=1.0,
        help="The default LR (1e-3 for Adam, 1.0 for SGD) is scaled by this factor.",
    )
    train_parser.add_argument(
        "--momentum",
        type=float,
        default=0.9,
        help="Momentum coefficient for SGD, or beta1 for Adam.",
    )
    train_parser.add_argument(
        "--num-steps", type=int, default=250, help="Number of training steps."
    )
    train_parser.add_argument(
        "--optimizer",
        type=str,
        default="sgd",
        choices=("adam", "sgd"),
        help="The type of optimizer to use.",
    )
    train_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=True,
        help="File to save the lenses to. Defaults to the model name.",
    )
    train_parser.add_argument(
        "--pre-ln",
        action="store_true",
        help="Apply layer norm before, and not after, each probe.",
    )
    train_parser.add_argument(
        "--resume", type=Path, help="File to resume training from."
    )
    train_parser.add_argument(
        "--separate-unembeddings",
        action="store_true",
        help="Learn a separate unembedding for each layer.",
    )
    train_parser.add_argument(
        "--tokens-per-step",
        type=int,
        default=2**18,
        help="Number of tokens per step.",
    )
    train_parser.add_argument(
        "--wandb", type=str, help="Name of run in Weights & Biases."
    )
    train_parser.add_argument(
        "--warmup-steps",
        type=int,
        default=None,
        help="Number of warmup steps. Defaults to min(0.1 * num_steps, 1000) for Adam"
        " and 0 for SGD.",
    )
    train_parser.add_argument(
        "--weight-decay", type=float, default=1e-3, help="Weight decay coefficient."
    )
    train_parser.add_argument(
        "--zero", action="store_true", help="Use ZeroRedundancyOptimizer."
    )

    downstream_parser.add_argument(
        "--lens",
        type=Path,
        help="Directory containing the tuned lens to evaluate.",
        nargs="?",
    )
    downstream_parser.add_argument(
        "--injection", action="store_true", help="Simulate a prompt injection attack."
    )
    downstream_parser.add_argument(
        "--incorrect-fewshot", action="store_true", help="Permute the fewshot labels."
    )
    downstream_parser.add_argument(
        "--num-shots",
        type=int,
        default=0,
        help="Number of examples to use for few-shot evaluation.",
    )
    downstream_parser.add_argument(
        "--limit", type=int, default=500, help="Number of samples to evaluate on."
    )
    downstream_parser.add_argument(
        "-o", "--output", type=Path, help="Folder to save the results to."
    )

    # Evaluation-only arguments
    eval_parser.add_argument(
        "--lens",
        type=Path,
        help="Directory containing the tuned lens to evaluate.",
        nargs="?",
    )
    eval_parser.add_argument(
        "--grad-alignment", action="store_true", help="Evaluate gradient alignment."
    )
    eval_parser.add_argument(
        "--limit",
        type=int,
        default=None,
        help="Number of batches to evaluate on. If None, will use the entire dataset.",
    )
    eval_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="JSON file to save the eval results to.",
    )
    eval_parser.add_argument(
        "--transfer",
        action="store_true",
        help="Evaluate how well probes transfer to other layers.",
    )

    args = parser.parse_args()

    # Support both distributed and non-distributed training
    local_rank = os.environ.get("LOCAL_RANK")
    if local_rank is not None:
        dist.init_process_group("nccl")
        local_rank = int(local_rank)

    # Only print on rank 0
    with nullcontext() if not local_rank else redirect_stdout(None):
        args = parser.parse_args()

        if args.sweep:
            ckpt_range = eval(f"range({args.sweep})")
            output_root = args.output
            assert output_root is not None
            print(f"Running sweep over {len(ckpt_range)} checkpoints.")

            for step in ckpt_range:
                step_output = output_root / f"step{step}"
                print(f"Running for step {step}, saving to '{step_output}'...")

                args.output = step_output
                args.revision = f"step{step}"
                lens_main(args)
        else:
            lens_main(args)


if __name__ == "__main__":
    run()



from tuned_lens.nn import Decoder
import transformers as tr
import pytest
import torch as th


def get_final_layer_norm(model: tr.AutoModelForCausalLM):
    """Get the final layer norm from a model.

    This isn't standardized across models, so this will need to be updated
    """
    base_model = model.base_model
    if isinstance(base_model, tr.models.opt.modeling_opt.OPTModel):
        return base_model.decoder.final_layer_norm
    elif isinstance(base_model, tr.models.gpt_neox.modeling_gpt_neox.GPTNeoXModel):
        return base_model.final_layer_norm
    elif isinstance(
        base_model,
        (
            tr.models.bloom.modeling_bloom.BloomModel,
            tr.models.gpt2.modeling_gpt2.GPT2Model,
            tr.models.gpt_neo.modeling_gpt_neo.GPTNeoModel,
            tr.models.gptj.modeling_gptj.GPTJModel,
        ),
    ):
        return base_model.ln_f
    else:
        raise NotImplementedError(f"Unknown model type {type(base_model)}")


def correctness(model_str: str):
    th.manual_seed(42)

    # We use a random model with the correct config instead of downloading the
    # whole pretrained checkpoint.
    config = tr.AutoConfig.from_pretrained(model_str)
    model = tr.AutoModelForCausalLM.from_config(config)

    # One problem: we want to check that we handle GPT-J's unembedding bias
    # correctly, but it's zero-initialized. Give it a random Gaussian bias.
    U = model.get_output_embeddings()
    if U.bias is not None:
        U.bias.data.normal_()

    decoder = Decoder(model)
    ln_f = get_final_layer_norm(model)

    x = th.randn(1, 1, config.hidden_size)
    y = U(ln_f(x)).log_softmax(-1)  # type: ignore[attr-defined]

    th.testing.assert_close(y, decoder(x).log_softmax(-1))

    x_hat = decoder.back_translate(x, tol=1e-5)
    th.testing.assert_close(y.exp(), decoder(x_hat).softmax(-1), atol=5e-4, rtol=0.01)


@pytest.mark.slow
def test_correctness_slow():
    correctness("EleutherAI/gpt-j-6B")


@pytest.mark.parametrize(
    "model_str",
    [
        "EleutherAI/pythia-125m",
        "bigscience/bloom-560m",
        "EleutherAI/gpt-neo-125M",
        "facebook/opt-125m",
        "gpt2",
    ],
)
def test_correctness(model_str: str):
    correctness(model_str)



from tuned_lens.nn.lenses import TunedLens, LogitLens
from transformers import AutoModelForCausalLM, AutoTokenizer
from tuned_lens.plotting import plot_lens
import torch as th


def test_plot_logit_lens():
    pythia_125M_model = AutoModelForCausalLM.from_pretrained("EleutherAI/pythia-125m")
    tokenizer = AutoTokenizer.from_pretrained("EleutherAI/pythia-125m")
    pythia_125M_lens = TunedLens(pythia_125M_model)
    pythia_125M_lens.attn_translators = th.nn.ModuleList()
    text = "Never gonna give you up, never gonna let you down,"
    text += " never gonna run around and desert you"
    input_ids = tokenizer(text, return_tensors="pt").input_ids

    # plot logit lens
    plot_lens(
        pythia_125M_model,
        lens=LogitLens(pythia_125M_model),
        input_ids=input_ids,
        tokenizer=tokenizer,
    )
    # plot w/ tuned lens
    plot_lens(
        pythia_125M_model,
        input_ids=input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
    )
    # plot w/ tuned lens and last layer output
    plot_lens(
        pythia_125M_model,
        input_ids=input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
    )

    # Plot w/ short text
    short_text = "Never gonna"
    short_input_ids = tokenizer(short_text, return_tensors="pt").input_ids
    plot_lens(
        pythia_125M_model,
        input_ids=short_input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
    )

    # Plot topk prob diff w/ last layer, should have N/A for the bottom row
    plot_lens(
        pythia_125M_model,
        input_ids=input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
        topk_diff=True,
    )
    # Plot topk prob diff w/o last layer
    plot_lens(
        pythia_125M_model,
        input_ids=input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
        topk_diff=True,
    )

    # Plot w/ topk equals large
    plot_lens(
        pythia_125M_model,
        input_ids=input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
        topk=30,
    )  # Breaks at 18 for notebook, but good for browser
    # Plot w/ topk equals small
    plot_lens(
        pythia_125M_model,
        input_ids=input_ids,
        tokenizer=tokenizer,
        lens=pythia_125M_lens,
        topk=1,
    )



from itertools import islice
from typing import cast, Any, Callable, Iterable, Sequence, Type, TypeVar, Union
import torch as th
import torch.distributed as dist


T = TypeVar("T")


def assert_type(typ: Type[T], obj: Any) -> T:
    """Assert that an object is of a given type at runtime and return it."""
    if not isinstance(obj, typ):
        raise TypeError(f"Expected {typ.__name__}, got {type(obj).__name__}")

    return cast(typ, obj)


def maybe_all_cat(x: th.Tensor) -> th.Tensor:
    if not dist.is_initialized():
        return x

    buffer = x.new_empty([dist.get_world_size() * x.shape[0], *x.shape[1:]])
    dist.all_gather_into_tensor(buffer, x)
    return buffer


def maybe_all_gather_lists(lst: list) -> list:
    if not dist.is_initialized():
        return lst

    lists = [[] for _ in range(dist.get_world_size())]
    dist.all_gather_object(lists, lst)
    return sum(lists, [])


def maybe_all_reduce(x: th.Tensor, op: str = "mean") -> th.Tensor:
    if not dist.is_initialized():
        return x

    if op == "sum":
        dist.all_reduce(x, op=dist.ReduceOp.SUM)
    elif op == "mean":
        dist.all_reduce(x, op=dist.ReduceOp.SUM)
        x /= dist.get_world_size()
    else:
        raise ValueError(f"Unknown reduction op '{op}'")

    return x


def maybe_unpack(x):
    """Unpack a tuple if it's a tuple, otherwise return the value."""
    if isinstance(x, tuple):
        x, *_ = x

    return x


def maybe_shift_labels(x: th.Tensor, shift: int):
    if shift > 0:
        return x[:, shift:]
    if shift < 0:
        return x[:, :shift]

    return x


def maybe_shift_preds(x: th.Tensor, shift: int):
    if shift > 0:
        return x[:, :-shift]
    if shift < 0:
        return x[:, -shift:]

    return x


T = TypeVar("T")


# Backported from Python 3.10
def pairwise(it: Iterable[T]) -> Iterable[tuple[T, T]]:
    """Iterate over pairs of elements in an iterable."""
    yield from zip(it, islice(it, 1, None))


# Define pytree type recursively- this works for Pylance but unfortunately not MyPy
AnyTree = Union[th.Tensor, dict[Any, "AnyTree"], list["AnyTree"], tuple["AnyTree", ...]]
TreeType = TypeVar("TreeType", bound=AnyTree)


def pytree_flatten(tree: AnyTree) -> Iterable[th.Tensor]:
    """Recursively iterate over all tensors in a pytree, in topological order."""
    # Stopping condition
    if isinstance(tree, th.Tensor):
        yield tree

    # Recursive case
    elif isinstance(tree, dict):
        for elem in tree.values():
            yield from pytree_flatten(elem)

    elif isinstance(tree, Sequence):
        for elem in tree:
            yield from pytree_flatten(elem)


def pytree_map(
    func: Callable[[th.Tensor], Any], tree: TreeType, strict: bool = True
) -> TreeType:
    """
    Recursively apply a function to all tensors in a pytree, returning the results
    in a new pytree with the same structure. Non-tensor leaves are copied.
    """
    # Stopping condition
    if isinstance(tree, th.Tensor):
        return func(tree)

    # Recursive case
    if isinstance(tree, dict):
        return {k: pytree_map(func, v) for k, v in tree.items()}

    if isinstance(tree, list):
        return [pytree_map(func, v) for v in tree]

    if isinstance(tree, tuple):
        return tuple(pytree_map(func, v) for v in tree)

    if strict:
        raise TypeError(
            f"Found leaf '{tree}' of unsupported type '{type(tree).__name__}'- use "
            f"`strict=False` to ignore"
        )
    else:
        return tree


def pytree_cat(trees: Sequence, dim: int = 0) -> AnyTree:
    """
    Concatenate pytrees along a given dimension, returning a new pytree with the same
    structure. All pytrees are expected to have the same structure; undefined behavior
    will occur if this is not the case.
    """
    transposed_iter = zip(*(pytree_flatten(tree) for tree in trees))
    leaf_iter = (th.cat(seq, dim) for seq in transposed_iter)
    try:
        return pytree_map(lambda _: next(leaf_iter), trees[0])  # type: ignore
    except (RuntimeError, StopIteration) as e:
        # Calling next() on an exhausted generator raises a RuntimeError, annoyingly
        if isinstance(e, StopIteration) or "StopIteration" in str(e):
            raise TypeError("All pytrees must have the same structure") from e
        else:
            raise


def pytree_stack(trees: Sequence, dim: int = 0) -> AnyTree:
    """
    Stack pytrees along a given dimension, returning a new pytree with the same
    structure. All pytrees are expected to have the same structure; undefined behavior
    will occur if this is not the case.
    """
    if not len(trees):
        raise ValueError("Cannot stack empty sequence of pytrees")

    transposed_iter = zip(*(pytree_flatten(tree) for tree in trees))
    leaf_iter = (th.stack(seq, dim) for seq in transposed_iter)
    try:
        return pytree_map(lambda _: next(leaf_iter), trees[0])  # type: ignore
    except (RuntimeError, StopIteration) as e:
        # Calling next() on an exhausted generator raises a RuntimeError, annoyingly
        if isinstance(e, StopIteration) or "StopIteration" in str(e):
            raise TypeError("All pytrees must have the same structure") from e
        else:
            raise


def revcumsum(x: Sequence[th.Tensor]) -> list[th.Tensor]:
    """Reverse cumulative sum of a sequence of tensors."""
    if not len(x):
        return []

    running_total = th.zeros_like(x[0])
    sums = [running_total.add_(r).clone() for r in reversed(x)]
    sums.reverse()
    return sums


def send_to_device(tree: TreeType, device: th.device) -> TreeType:
    """Recursively send all tensors in a pytree to a device."""
    return pytree_map(lambda t: t.to(device), tree)



from torch.distributions import Categorical, Dirichlet, kl_divergence
from tuned_lens.stats import aitchison, aitchison_similarity, js_divergence, js_distance
import torch as th


def test_aitchison():
    # Uniform and symmetric Dirichlet prior over the simplex
    N, K = 100, 50_000
    prior = Dirichlet(th.ones(K, dtype=th.float64))

    log_x = prior.sample(th.Size([N])).log()
    log_y = prior.sample(th.Size([N])).log()
    log_z = prior.sample(th.Size([N])).log()
    weights = prior.sample(th.Size([N]))

    # Linearity:
    # <ax + by, z> = a<x, z> + b<y, z>
    a = th.randn(N, dtype=th.float64)
    b = th.randn(N, dtype=th.float64)

    th.testing.assert_close(
        # <ax + by, z>
        aitchison(
            th.log_softmax(a[:, None] * log_x + b[:, None] * log_y, dim=-1),
            log_z,
            weight=weights,
        ),
        # a<x, z> + b<y, z>
        (
            a * aitchison(log_x, log_z, weight=weights)
            + b * aitchison(log_y, log_z, weight=weights)
        ),
    )

    # Positive definiteness
    assert th.all(aitchison(log_x, log_x, weight=weights) >= 0)

    # Symmetry
    th.testing.assert_close(
        aitchison(log_x, log_y, weight=weights), aitchison(log_y, log_x, weight=weights)
    )

    # Cosine similarity properties
    self_similarities = aitchison_similarity(log_x, log_x, weight=weights)
    similarities = aitchison_similarity(log_x, log_y, weight=weights)
    assert th.all(similarities >= -1) and th.all(similarities <= 1)
    th.testing.assert_close(self_similarities, th.ones_like(self_similarities))


def test_js_divergence():
    p = Categorical(logits=th.randn(10))
    q = Categorical(logits=th.randn(10))
    m = Categorical(probs=0.5 * (p.probs + q.probs))  # type: ignore

    kl_fwd = kl_divergence(p, m)
    kl_bwd = kl_divergence(q, m)
    gt_js = 0.5 * (kl_fwd + kl_bwd)

    our_js_fwd = js_divergence(p.logits, q.logits)  # type: ignore
    our_js_bwd = js_divergence(q.logits, p.logits)  # type: ignore

    th.testing.assert_close(gt_js, our_js_fwd)
    th.testing.assert_close(our_js_fwd, our_js_bwd)  # Symmetry


def test_js_distance():
    a = th.randn(1000, 3)
    b = th.randn(1000, 3)
    c = th.randn(1000, 3)

    dist_ab = js_distance(a, b)
    dist_bc = js_distance(b, c)
    dist_ac = js_distance(a, c)

    # Triangle inequality
    assert th.all(dist_ab + dist_bc >= dist_ac)



from contextlib import contextmanager
from transformers import PreTrainedModel
from typing import Any, Generator, Optional, Type, TypeVar, Union
import torch as th


def get_value_for_key(obj: Any, key: str) -> Any:
    """Get a value using `__getitem__` if `key` is numeric and `getattr` otherwise."""
    return obj[int(key)] if key.isdigit() else getattr(obj, key)


def set_value_for_key_(obj: Any, key: str, value: Any) -> None:
    """
    Set value in-place using `__getitem__` if `key` is numeric and `getattr` otherwise.
    """
    if key.isdigit():
        obj[int(key)] = value
    else:
        setattr(obj, key, value)


def get_key_path(model: th.nn.Module, key_path: str) -> Any:
    """Get a value by key path, e.g. `layers.0.attention.query.weight`."""
    for key in key_path.split("."):
        model = get_value_for_key(model, key)

    return model


def set_key_path_(
    model: th.nn.Module, key_path: str, value: Union[th.nn.Module, th.Tensor]
) -> None:
    """Set a value by key path in-place, e.g. `layers.0.attention.query.weight`."""
    keys = key_path.split(".")
    for key in keys[:-1]:
        model = get_value_for_key(model, key)

    setattr(model, keys[-1], value)


T = TypeVar("T", bound=th.nn.Module)


@contextmanager
def assign_key_path(model: T, key_path: str, value: Any) -> Generator[T, None, None]:
    """Temporarily set a value by key path while in the context."""
    old_value = get_key_path(model, key_path)
    set_key_path_(model, key_path, value)
    try:
        yield model
    finally:
        set_key_path_(model, key_path, old_value)


def get_final_layer_norm(
    model: PreTrainedModel, norm_class: Type[th.nn.Module] = th.nn.LayerNorm
) -> Optional[th.nn.Module]:
    """Use heuristics to find the final layer norm in a model, if it exists."""
    base = model.base_model
    if decoder := getattr(base, "decoder", None):
        base = decoder

    top_level_lns = [
        module for module in base.children() if isinstance(module, norm_class)
    ]
    return top_level_lns[-1] if top_level_lns else None


def get_transformer_layers(model: th.nn.Module) -> tuple[str, th.nn.ModuleList]:
    """Get "the" list of transformer layers from a model.

    This is operationalized as the unique `nn.ModuleList` that contains
    more than half of all the parameters in the model, if it exists.

    Args:
        model: The model to search.

    Returns:
        A tuple containing the key path to the layer list and the list itself.

    Raises:
        ValueError: If no such list exists.
    """
    total_params = sum(p.numel() for p in model.parameters())
    for name, module in model.named_modules():
        if isinstance(module, th.nn.ModuleList):
            module_params = sum(p.numel() for p in module.parameters())
            if module_params > total_params / 2:
                return name, module

    raise ValueError(
        "Could not find suitable `ModuleList`; is this an encoder-decoder model?"
    )


@contextmanager
def delete_layers(model: T, indices: list[int]) -> Generator[T, None, None]:
    """Temporarily delete the layers at `indices` from `model` while in the context."""
    list_path, layer_list = get_transformer_layers(model)
    modified_list = th.nn.ModuleList(layer_list)
    for i in sorted(indices, reverse=True):
        del modified_list[i]

    set_key_path_(model, list_path, modified_list)
    try:
        yield model
    finally:
        set_key_path_(model, list_path, layer_list)


@contextmanager
def permute_layers(model: T, indices: list[int]) -> Generator[T, None, None]:
    """Temporarily permute the layers of `model` by `indices` while in the context.

    The number of indices provided may be not be equal to the number of
    layers in the model. Layers will be dropped or duplicated accordingly.
    """
    list_path, layer_list = get_transformer_layers(model)
    permuted_list = th.nn.ModuleList([layer_list[i] for i in indices])
    set_key_path_(model, list_path, permuted_list)

    try:
        yield model
    finally:
        set_key_path_(model, list_path, layer_list)


def permute_layers_(model: th.nn.Module, indices: list[int]):
    """Permute the layers of `model` by `indices` in-place.

    The number of indices provided may be not be equal to the number of
    layers in the model. Layers will be dropped or duplicated accordingly.
    """
    list_path, layer_list = get_transformer_layers(model)
    permuted_list = th.nn.ModuleList([layer_list[i] for i in indices])
    set_key_path_(model, list_path, permuted_list)


@contextmanager
def replace_layers(
    model: T, indices: list[int], replacements: list[th.nn.Module]
) -> Generator[T, None, None]:
    """Replace the layers at `indices` with `replacements` while in the context."""
    list_path, layer_list = get_transformer_layers(model)
    modified_list = th.nn.ModuleList(layer_list)
    for i, replacement in zip(indices, replacements):
        modified_list[i] = replacement

    set_key_path_(model, list_path, modified_list)
    try:
        yield model
    finally:
        set_key_path_(model, list_path, layer_list)



import pytest
from tuned_lens.load_artifacts import load_lens_artifacts


@pytest.mark.skip(reason="The lens space needs to be made public first")
def test_load_lens_artifact_smoke():
    load_lens_artifacts("gpt2", "AlignmentResearch/tuned-lens")



from datasets import Dataset, DatasetDict
from functools import partial
from multiprocessing import cpu_count
from transformers import PreTrainedTokenizerBase
from typing import TypeVar, Union
import logging
import math
import warnings


T = TypeVar("T", bound=Union[Dataset, DatasetDict])


def chunk_and_tokenize(
    data: T,
    tokenizer: PreTrainedTokenizerBase,
    *,
    format: str = "torch",
    num_proc: int = cpu_count() // 2,
    text_key: str = "text",
) -> T:
    """Perform GPT-style chunking and tokenization on a dataset.

    The resulting dataset will consist entirely of chunks exactly `max_length` tokens
    long. Long sequences will be split into multiple chunks, and short sequences will
    be merged with their neighbors, using `eos_token` as a separator.

    Args:
        data: The dataset to chunk and tokenize.
        tokenizer: The tokenizer to use.
        format: The format to return the dataset in, passed to `Dataset.with_format`.
        num_proc: The number of processes to use for tokenization.
        text_key: The key in the dataset to use as the text to tokenize.

    Returns:
        The chunked and tokenized dataset.
    """
    return data.map(
        partial(_tokenize_fn, tokenizer=tokenizer, text_key=text_key),
        batched=True,
        num_proc=num_proc,
        remove_columns=get_columns_all_equal(data),
    ).with_format(
        format,
        # Remove the "overflow_to_sample_mapping" column so we can directly pass
        # elements of the dataset to a model
        columns=["input_ids"],
    )


def compute_nats_to_bpb_ratio(raw: T, tokenized: T) -> float:
    """Compute ratio of nats per token to bits per byte for a given tokenizer.

    This is used to convert the perplexity of a model to bits per byte.

    Args:
        raw: The raw, unprocessed dataset.
        tokenized: The tokenized dataset.

    Returns:
        The ratio of nats to bits per byte.
    """
    byte_counts = raw.map(
        lambda x: {"length": [len(txt.encode("utf-8")) for txt in x["text"]]},
        batched=True,
        num_proc=cpu_count() // 2,
        remove_columns=get_columns_all_equal(raw),
    )

    token_counts = tokenized.map(
        lambda x: {"length": [len(ids) for ids in x["input_ids"]]},
        batched=True,
        num_proc=cpu_count() // 2,
        remove_columns=get_columns_all_equal(tokenized),
    )
    total_bytes = sum(byte_counts["length"])  # type: ignore[operator]
    total_tokens = sum(token_counts["length"])  # type: ignore[operator]

    # See https://arxiv.org/pdf/2101.00027.pdf, section 3.1
    return (total_tokens / total_bytes) / math.log(2)


def _tokenize_fn(x: dict[str, list], tokenizer: PreTrainedTokenizerBase, text_key: str):
    """Annoyingly, we need to use a separate function so it can be hashed correctly."""
    chunk_size = min(tokenizer.model_max_length, 2048)
    sep = tokenizer.eos_token or "<|endoftext|>"
    output = tokenizer(
        # Concatenate all the samples together, separated by the EOS token.
        sep.join(x[text_key]),
        max_length=chunk_size,
        return_overflowing_tokens=True,
        truncation=True,
    )

    # Workaround for weird inconsistency of behavior in OPT tokenizer
    if overflow := output.pop("overflowing_tokens", None):
        tokens = output["input_ids"]
        tokens += overflow

        # Drop tokens that don't fit into a chunk
        limit = len(tokens) - len(tokens) % chunk_size

        # Manually group tokens into a list of chunks of length `chunk_size`
        return {
            "input_ids": [
                tokens[i : i + chunk_size] for i in range(0, limit, chunk_size)
            ]
        }

    # Normal case
    return {
        # We know that the last sample will almost always be less than the max
        # number of tokens, and we don't want to pad, so we just drop it.
        k: v[:-1]
        for k, v in output.items()
    }


def get_columns_all_equal(dataset: Union[Dataset, DatasetDict]) -> list[str]:
    """Get a single list of columns in a `Dataset` or `DatasetDict`.

    We assert the columms are the same across splits if it's a `DatasetDict`.

    Args:
        dataset: The dataset to get the columns from.

    Returns:
        A list of columns.
    """
    if isinstance(dataset, DatasetDict):
        cols_by_split = dataset.column_names.values()
        columns = next(iter(cols_by_split))
        if not all(cols == columns for cols in cols_by_split):
            raise ValueError("All splits must have the same columns")

        return columns

    return dataset.column_names


def silence_datasets_messages():
    """Silence the annoying wall of logging messages and warnings."""

    def filter_fn(log_record):
        msg = log_record.getMessage()
        return (
            "Found cached dataset" not in msg
            and "Loading cached" not in msg
            and "Using custom data configuration" not in msg
        )

    handler = logging.StreamHandler()
    handler.addFilter(filter_fn)
    logging.getLogger("datasets").addHandler(handler)

    warnings.filterwarnings("ignore", category=FutureWarning, module="datasets")



from typing import Optional, Tuple
from pathlib import Path
import os

from huggingface_hub import hf_hub_download


def load_lens_artifacts(
    resource_id: str,
    repo_id: Optional[str] = None,
    repo_type: Optional[str] = None,
    revision: Optional[str] = None,
    config_file: str = "config.json",
    ckpt_file: str = "params.pt",
    subfolder: str = "lens",
) -> Tuple[Path, Path]:
    """First checks for lens resource locally then tries to download it from the hub

    Args:
        resource_id: The id of the lens resource.
        repo_id: The repository to download the lens from. Defaults to
            'AlignmentResearch/tuned-lens'. However, this default can be overridden by
            setting the TUNED_LENS_REPO_ID environment variable.
        repo_type: The type of repository to download the lens from. Defaults to
            'space'. However, this default can be overridden by setting the
            TUNED_LENS_REPO_TYPE environment variable.
        revision: The revision of the lens to download.
        subfolder: The subfolder of the repository to download the lens from.

    Returns:
        * The path to the config.json file
        * The path to the params.pt file

    Raises:
        ValueError: if the lens resource could not be found.
    """
    if repo_id is None:
        if os.environ.get("TUNED_LENS_REPO_ID"):
            repo_id = os.environ["TUNED_LENS_REPO_ID"]
        else:
            repo_id = "AlignmentResearch/tuned-lens"

    if repo_type is None:
        if os.environ.get("TUNED_LENS_REPO_TYPE"):
            repo_type = os.environ["TUNED_LENS_REPO_TYPE"]
        else:
            repo_type = "space"

    # Fist check if the resource id is a path to a folder that exists
    local_path = Path(resource_id)
    if (local_path / config_file).exists() and (local_path / ckpt_file).exists():
        return local_path / config_file, local_path / ckpt_file

    subfolder = "/".join((subfolder, resource_id))
    params_path = hf_hub_download(
        filename=ckpt_file,
        repo_id=repo_id,
        repo_type=repo_type,
        revision=revision,
        subfolder=subfolder,
    )

    config_path = hf_hub_download(
        filename=config_file,
        repo_id=repo_id,
        repo_type=repo_type,
        revision=revision,
        subfolder=subfolder,
    )

    if config_path is not None and params_path is not None:
        return Path(config_path), Path(params_path)

    raise ValueError("Could not find lens resource locally or on the hf hub.")
